#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <strings.h>
#include <pthread.h>
#include <errno.h>
#include <omp.h>

void driverthread()
{
  int s;
  struct sockaddr_in addr;
  bzero( &addr, sizeof(addr) );
  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = inet_addr( "127.0.0.1" );
  addr.sin_port = htons( 5000 );

  #pragma omp parallel for shared( addr ) private( s )
  for( int count=0; count<10000; count++ )
  {
    s = socket( PF_INET, SOCK_STREAM, 0 );
    printf( "Driver thread %i ready\n", omp_get_thread_num() );

    if ( connect( s, (struct sockaddr*)&addr, sizeof(addr) )==0 )
    {
      char buffer[1024];

      for ( int i=0; i<10; i++ )
      {
        sprintf( buffer, "Sent %i\n", i );
        if ( send( s, buffer, strlen(buffer)+1 ,0 )!= strlen(buffer)+1 )
        { 
           printf( "send size mismatch\n" ); 
        }
        bzero( buffer, sizeof(buffer) );
        read(s, buffer, sizeof(buffer) );
      }

    }
    else
    {
      perror( "Connection refused" );
      exit( 0 );
    }

   shutdown( s, SHUT_RDWR );
   close( s );
  }
}

void echothread()
{
  int s;
  int true = 1;
  struct sockaddr_in addr;

  s = socket( PF_INET, SOCK_STREAM, 0 );
  if ( s == -1 ) { printf( "Socket error %i\n", errno ); }

  if ( setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &true, sizeof(int))==-1 )
     { printf( "setsockopt error %i\n", errno ); }

  bzero( &addr, sizeof(addr) );
  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = htonl( INADDR_ANY );
  addr.sin_port = htons( 5000 );

  if ( bind( s, (struct sockaddr*)&addr, sizeof(addr) ) != 0 )
     { printf( "Bind error %i\n", errno ); }

  listen( s, 4 );
  #pragma omp parallel
  {
    #pragma omp single
    while( 1 )
    {
      struct sockaddr client;

      int size = sizeof(client);
      int stream = accept( s, &client, &size );
      #pragma omp task
      {
        char buffer[1024];

        if ( stream >= 0 ) 
         { printf("Accepted by thread ID %i\n", omp_get_thread_num()); }
        else 
         { printf("Accept error %i\n", errno); }

        while ( recv( stream, buffer, sizeof(buffer), 0 ) )
        {
          send( stream, buffer, strlen(buffer)+1, 0 );
        }

        close( stream );
        printf( "Stream closed\n" );
      }
    }
  }
}


void main()
{
  omp_set_nested( 1 );
  #pragma omp parallel sections num_threads( 2 )
  {
    #pragma omp section
    {
      echothread();
    }
    #pragma omp section
    {
      driverthread();
    }
  }
}

