#include <stdio.h>
#include <pthread.h>

volatile int volatile buffer[16];
volatile int addhere;
volatile int removehere;
volatile int lock = 0;

#ifdef __sparc
int CAS( volatile int* addr, int ov, int nv)
{
  asm volatile( "cas %1, %2, %0": 
                "=r"(nv): 
                "m"(*addr),"r"(ov),"0"(nv):
                "memory" );
  return nv;
}
#else
int CAS( volatile int* addr, int ov, int nv )
{
  asm volatile( "lock; cmpxchg %2, %1":
                "=a"(ov):
                "m"(*addr),"r"(nv),"a"(ov):
                "memory" );
  return ov;
}
#endif

void lock_spinlock( volatile int *  lock )
{
  int acquired = 0;
  while ( !acquired ) 
  { 
    if ( ( *lock == 0 ) && ( CAS( lock, 0, 1 ) == 0) ) 
    {
      acquired = 1;
    }
  }
}

void free_spinlock( volatile int * lock )
{
  *lock = 0;
}

void clearbuffer()
{
  addhere = 0;
  removehere = 0;
  for( int i=0; i<16; i++ ) { buffer[i] = 0; }
}

int next( int current )
{
  return ( current + 1 ) & 15;
}

void addtobuffer( int value )
{
  int success = 0;
  while ( !success )
  {
    if ( buffer[ next(addhere) ] == 0 )
    {
      lock_spinlock( &lock );
      if ( buffer[ next(addhere) ] == 0 )
      {
        buffer[addhere] = value;
        addhere = next(addhere);
        success = 1;
      }
      free_spinlock( &lock );
    }
  }
}

int removefrombuffer()
{
  int value;
  int success = 0;
  while ( !success )
  {
    if ( buffer[removehere] != 0 )
    {
      lock_spinlock( &lock );
      if ( ( value = buffer[removehere] ) !=0 )
      {
        buffer[removehere] = 0;
        removehere = next(removehere);
        success = 1;
      }
      free_spinlock( &lock );
    }
  }
  return value;
}

void * producer( void *param )
{
  for( int i=1; i<10000000; i++ )
  {
    addtobuffer( i );
  }
}

void * consumer ( void *param )
{
  while ( removefrombuffer() != 9999999 ) {}
}

int main()
{
  clearbuffer();
  pthread_t threads[2];
  pthread_create( &threads[0], 0, producer, 0 );
  pthread_create( &threads[1], 0, consumer, 0 );
  pthread_join( threads[1], 0 );
  pthread_join( threads[0], 0 );
}

